"""Type supporting native Python3 enum. It depends either on Py3.4+ or e.g. enum34.
"""

from enum import Enum

from ..exceptions import ConversionError
from ..translator import _
from ..types import BaseType


class EnumType(BaseType):
    """A field type allowing to use native enums as values.
    Restricts values to enum members and (optionally) enum values.
    `use_values` - if set to True allows do assign enumerated values to the field.

    >>> import enum
    >>> class E(enum.Enum):
    ...    A = 1
    ...    B = 2
    >>> from schematics import Model
    >>> class AModel(Model):
    ...    foo = EnumType(E)
    >>> a = AModel()
    >>> a.foo = E.A
    >>> a.foo.value == 1
    """

    MESSAGES = {
        "convert": _("Couldn't interpret '{0}' as member of {1}."),
    }

    def __init__(self, enum, use_values=False, **kwargs):
        """
        :param enum: Enum class to which restrict values assigned to the field.
        :param use_values: If true, also values of the enum (right-hand side) can be assigned here.
        Other args are passed to superclass.
        """
        self._enum_class = enum
        self._use_values = use_values
        super().__init__(**kwargs)

    def to_native(self, value, context=None):
        if isinstance(value, self._enum_class):
            return value
        by_name = self._find_by_name(value)
        if by_name:
            return by_name
        by_value = self._find_by_value(value)
        if by_value:
            return by_value
        raise ConversionError(self.messages["convert"].format(value, self._enum_class))

    def _find_by_name(self, value):
        if isinstance(value, str):
            try:
                return self._enum_class[value]
            except KeyError:
                pass
        return None

    def _find_by_value(self, value):
        if not self._use_values:
            return None
        for member in self._enum_class:
            if member.value == value:
                return member
        return None

    def to_primitive(self, value, context=None):
        if isinstance(value, Enum):
            if self._use_values:
                return value.value
            return value.name
        return str(value)



from copy import deepcopy

import pytest

from schematics.exceptions import *
from schematics.models import Model
from schematics.types import *
from schematics.types.compound import *
from schematics.undefined import Undefined


@pytest.mark.parametrize("init", (True, False))
def test_import_data(init):
    class M(Model):
        a, b, c, d = IntType(), IntType(), IntType(), IntType()

    m = M({"a": 1, "b": None, "c": 3}, init=init)

    m.import_data({"a": None, "b": 2})

    if init:
        assert m._data == {"a": None, "b": 2, "c": 3, "d": None}
    else:
        assert m._data == {"a": None, "b": 2, "c": 3}


@pytest.mark.parametrize("init", (True, False))
def test_import_data_with_error(init):
    class M(Model):
        a, b, c, d = IntType(), IntType(), IntType(required=True), IntType()

    m = M({"a": 1, "b": None, "c": 3}, init=init)

    with pytest.raises(DataError):
        m.import_data(
            {
                "a": None,
                "b": 2,
                "c": None,
            }
        )

    if init:
        assert m._data == {"a": 1, "b": None, "c": 3, "d": None}
    else:
        assert m._data == {"a": 1, "b": None, "c": 3}


@pytest.mark.parametrize(
    "preconvert_source, populate_source", [(False, None), (True, True), (True, False)]
)
@pytest.mark.parametrize(
    "recursive, populate_target, init_to_none, populated_result",
    [
        (False, True, True, True),
        (False, False, False, False),
        (True, True, True, True),
        (True, False, True, True),
        (True, False, False, False),
    ],
)
def test_complex_import_data(
    recursive,
    preconvert_source,
    populate_source,
    populate_target,
    init_to_none,
    populated_result,
):
    class M(Model):
        intfield = IntType(max_value=2)
        matrixfield = ListType(ListType(IntType))
        dictfield = DictType(IntType)
        modelfield = ModelType("M")

    origdict = {
        "intfield": "1",
        "dictfield": dict(a=1, b=2),
        "modelfield": {
            "intfield": "2",
            "matrixfield": [[0, 0, 0], [1, 1, 1], [2, 2, 2]],
            "dictfield": dict(a=11, b=22),
            "modelfield": {"intfield": "3", "dictfield": dict(a=111, b=222)},
        },
    }

    m = M(origdict, init=populate_target)

    sourcedict = {
        "intfield": "101",
        "dictfield": dict(c=3),
        "modelfield": {
            "matrixfield": [[9]],
            "modelfield": {"intfield": "103", "dictfield": dict(c=33)},
        },
    }

    sourcedata = deepcopy(sourcedict)

    if preconvert_source:
        sourcedata = M(sourcedata, init=populate_source)

    m.import_data(sourcedata, recursive=recursive, init_values=init_to_none)

    assert id(m) != id(sourcedata)

    if preconvert_source and populate_source:
        assert m == M(sourcedict, init=True)
    elif recursive:
        assert m == M(
            {
                "intfield": "101",
                "dictfield": dict(c=3),
                "modelfield": {
                    "intfield": "2",
                    "matrixfield": [[9]],
                    "dictfield": dict(a=11, b=22),
                    "modelfield": {"intfield": "103", "dictfield": dict(c=33)},
                },
            },
            init=populated_result,
        )
    else:
        assert m == M(sourcedict, init=populated_result)
